function [varargout] = ml_hinfbt(varargin)
%ML_HINFBT H-infinity balanced truncation for standard systems.
%
% DESCRIPTION:
%   This function computes the H-infinity balanced truncation for a
%   standard system of the form
%
%       x'(t) = A*x(t) + B*u(t),                                        (1)
%        y(t) = C*x(t) + D*u(t).                                        (2)
%
%   Therefor, the two algebraic Riccati equations
%
%       A*P  + P*A' + B*B'
%           - beta * (P*C' + B*D') * inv(Rb) * (P*C' + B*D')' = 0,
%       A'*Q + Q*A  + C'*C
%           - beta * (B'*Q + D'*C)' * inv(Rc) * (B'*Q + D'*C) = 0,
%
%   are solved for the Gramians P and Q, with beta = (1 - GAM^(-2)) and
%
%       Rb = I + D*D',
%       Rc = I + D'*D.
%
%   As result, a reduced-order system of the form
%
%       x'(t) = Ar*x(t) + Br*u(t),                                      (3)
%        y(t) = Cr*x(t) + Dr*u(t)                                       (4)
%
%   is computed, such that for the original transfer function G with the
%   (right) coprime factorization beta*G = N*inv(M) and the r-th order
%   transfer function Gr with the (right) coprime factorization
%   beta*Gr = Nr*inv(Mr) it holds
%
%       1/beta * ||[N; M] - [Nr; Mr]||_{\infty} <= 2 * (Hsv(r+1)/sqrt(1
%       + (beta*Hsv(r+1))^2) + ... + Hsv(n)/sqrt(1 + (beta*Hsv(n))^2)),
%
%   with Hsv, a vector containing the characteristic H-infinity singular
%   values of the system.
%
%   Note:
%     1) The parameter GAM refers to the optimal cost gamma computed during
%        the H-infinity controller construction of the LQG normalized
%        system with feed-through term
%
%           [ A | B 0 B ]
%           [ --------- ]
%           [ C | 0 0 0 ],
%           [ 0 | 0 0 I ]
%           [ C | 0 I D ]
%
%        which can be constructed by
%
%           sys_normal = ss(A, ...
%                           [B, zeros(n,p), B], ...
%                           [C; zeros(m,n); C], ...
%                           [zeros(p,m), zeros(p,p), zeros(p,m);
%                            zeros(m,m), zeros(m,p), eye(m,m);
%                            zeros(p,m), eye(p,p), D]).
%
%        Then an optimal parameter GAM_OPT can be computed by
%
%           [K, CL, GAM_OPT, INFO] = hinfsyn(sys_normal, p, m).
%
%        In this case, the parameter GAM > GAM_OPT can be chosen
%        arbitrarily.
%        By default, the maximum of 1 and the given GAM is disturbed by
%        0.01% before used for computations.
%     2) In the case of GAM == Inf, the H-infinity balanced truncation is
%        identical to the LQG balanced truncation.
%
%
% 1) If the standard system is given by matrices:
% SIGNATURE:
%   [Ar, Br, Cr, Dr, info] = ML_HINFBT(A, B, C, D)
%   [Ar, Br, Cr, Dr, info] = ML_HINFBT(A, B, C, D, opts)
%
% INPUT:
%   A    - matrix from (1) with dimensions n x n
%   B    - matrix from (1) with dimensions n x m
%   C    - matrix from (2) with dimensions p x n
%   D    - matrix from (2) with dimensions p x m
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | Beta            | nonnegative scalar, used as shift of the          |
%   |                 | in Bass' algorithm for better conditioning if     |
%   |                 | StabMethod == 'lyap' is chosen                    |
%   |                 | (default 0.1)                                     |
%   |-----------------|---------------------------------------------------|
%   | careopts        | structure, containing the optional parameters for |
%   |                 | the computation of the continuous-time algebraic  |
%   |                 | Riccati equations, with                           |
%   |                 |   AbsTol   - tolerance for the absolute error     |
%   |                 |              (default 0)                          |
%   |                 |   Info     - disable/enable display of verbose    |
%   |                 |              status information during iteration  |
%   |                 |              (default 0)                          |
%   |                 |   lyapopts - structure, containing the optional   |
%   |                 |              parameters for the Lyapunov solver   |
%   |                 |              used in each iteration step          |
%   |                 |     * AbsTol  - tolerance for the absolute error  |
%   |                 |                 (default 0)                       |
%   |                 |     * CompTol - tolerance for column compression  |
%   |                 |                 (default log(n)*eps)              |
%   |                 |     * Info    - disable/enable display of verbose |
%   |                 |                 status information                |
%   |                 |                 (default 0)                       |
%   |                 |     * MaxIter - maximum number of iteration steps |
%   |                 |                 (default 100)                     |
%   |                 |     * RelTol  - tolerance for the relative error  |
%   |                 |                 (default 1.0e+01*n*eps)           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative error      |
%   |                 |             (default 1.0e+02*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | Gamma           | positive scalar, used for the scaling of the      |
%   |                 | Riccati equations, see GAM above                  |
%   |                 | (default Inf)                                     |
%   |-----------------|---------------------------------------------------|
%   | Method          | character array, determining algorithm for the    |
%   |                 | computation of the reduced-order model            |
%   |                 |  'sr'   - square-root method                      |
%   |                 |  'bfsr' - balancing-free square-root method       |
%   |                 | (default 'sr')                                    |
%   |-----------------|---------------------------------------------------|
%   | Order           | positive integer, order of the resulting          |
%   |                 | reduced-order model chosen by the user if         |
%   |                 | 'order' is set for OrderComputation               |
%   |                 | (default min(10,length(Hsv)) + Nu)                |
%   |-----------------|---------------------------------------------------|
%   | OrderComputation| character array, determining the method for the   |
%   |                 | computation of the size of the reduced-order model|
%   |                 |  'order'     - take explicit order                |
%   |                 |  'tolerance' - using absolute error bound         |
%   |                 | (default 'tolerance')                             |
%   |-----------------|---------------------------------------------------|
%   | StabMethod      | character array, determining the method of        |
%   |                 | stabilization of the system                       |
%   |                 |   'abe'  - partial stabilization with the         |
%   |                 |            algebraic Bernoulli equation           |
%   |                 |   'lyap' - partial stabilization with Bass'       |
%   |                 |            algorithm                              |
%   |                 | (default 'abe')                                   |
%   |-----------------|---------------------------------------------------|
%   | stabmethodopts  | structure, containing the optional parameters for |
%   |                 | the sign function based Lyapunov or Bernoulli     |
%   |                 | equation solver used for the stabilization, with  |
%   |                 |   AbsTol    - tolerance for the absolute error    |
%   |                 |               (default 0)                         |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RelTol    - tolerance for the relative error    |
%   |                 |               (default 1.0e+01*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | stabsignmopts   | structure, containing the optional parameters for |
%   |                 | the matrix sign function used for the             |
%   |                 | decomposition into stable and anti-stable system  |
%   |                 | parts                                             |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | Tolerance       | nonnegative scalar, tolerance used for the        |
%   |                 | computation of the size of the reduced-order model|
%   |                 | by an absolute error bound if 'tolerance' is set  |
%   |                 | for OrderComputation                              |
%   |                 | (default 1.0e-02)                                 |
%   |-----------------|---------------------------------------------------|
%   | UnstabDim       | integer, dimension of the deflating anti-stable   |
%   |                 | subspace in the control and filter Riccati        |
%   |                 | equations, negative if unknown                    |
%   |                 | (default -1)                                      |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   Ar   - matrix of (3) with dimensions r x r
%   Br   - matrix of (3) with dimensions r x m
%   Cr   - matrix of (4) with dimensions p x r
%   Dr   - matrix of (4) with dimensions p x m
%   info - structure, containing the following information:
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | AbsErrBound     | computed error bound for the absolute error of the|
%   |                 | (right) coprime factorization in H-infinity norm  |
%   |-----------------|---------------------------------------------------|
%   | Gamma           | positive scalar, the disturbed scaling parameter  |
%   |                 | of the computations                               |
%   |-----------------|---------------------------------------------------|
%   | Hsv             | a vector, containing the computed characteristic  |
%   |                 | H-infinity singular values                        |
%   |-----------------|---------------------------------------------------|
%   | infoCARE_C      | structure, containing information about the       |
%   |                 | continuous-time algebraic Riccati equation solver |
%   |                 | for the for the controllability Gramian, with     |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | infoCARE_O      | structure, containing information about the       |
%   |                 | continuous-time algebraic Riccati equation solver |
%   |                 | for the for the observability Gramian, with       |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | infoPARTSTAB_C  | structure, containing information about the       |
%   |                 | partial stabilization used for the controllability|
%   |                 | Riccati equation                                  |
%   |                 |   infoSTABMETH   - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    stabilization                  |
%   |                 |   infoSTABMETH2  - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    a second stabilization         |
%   |                 |   infoSTABSIGNM  - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    separation of anti-stable      |
%   |                 |                    eigenvalues                    |
%   |                 |   infoSTABSIGNM2 - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    a second separation of anti-   |
%   |                 |                    stable eigenvalues             |
%   |                 |   Method         - character array with shortcut  |
%   |                 |                    of the used method             |
%   |                 |   Ns             - identified stable eigenvalues  |
%   |                 |   Nu             - identified anti-stable         |
%   |                 |                    eigenvalues                    |
%   |-----------------|---------------------------------------------------|
%   | infoPARTSTAB_O  | structure, containing information about the       |
%   |                 | partial stabilization used for the observability  |
%   |                 | Riccati equation                                  |
%   |                 |   infoSTABMETH   - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    stabilization                  |
%   |                 |   infoSTABMETH2  - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    a second stabilization         |
%   |                 |   infoSTABSIGNM  - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    separation of anti-stable      |
%   |                 |                    eigenvalues                    |
%   |                 |   infoSTABSIGNM2 - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    a second separation of anti-   |
%   |                 |                    stable eigenvalues             |
%   |                 |   Method         - character array with shortcut  |
%   |                 |                    of the used method             |
%   |                 |   Ns             - identified stable eigenvalues  |
%   |                 |   Nu             - identified anti-stable         |
%   |                 |                    eigenvalues                    |
%   |-----------------|---------------------------------------------------|
%   | N               | Dimension of the reduced-order model              |
%   |-----------------|---------------------------------------------------|
%
% 2) If the standard system is given in a struct:
% SIGNATURE:
%   [rom, info] = ML_HINFBT(sys)
%   [rom, info] = ML_HINFBT(sys, opts)
%
% INPUT:
%   sys  - structure, containing the standard system's matrices
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (1) with dimensions n x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (2) with dimensions p x n             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (2) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - structure, with the following entries
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (3) with dimensions r x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (4) with dimensions p x r             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (4) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing information (see above 1))
%
% 3) If the standard system is given as state-space object:
% SIGNATURE:
%   [rom, info] = ML_HINFBT(sys)
%   [rom, info] = ML_HINFBT(sys, opts)
%
% INPUT:
%   sys  - state-space object, containing the standard system's matrices
%          (see above 2))
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - state-space object, containing the reduced-order standard system
%          (see above 2))
%   info - structure, containing information (see above 1))
%
%
% REFERENCE:
%   D. MUSTAFA, K. GLOVER. Controller reduction by Hinf-balanced
%   truncation. IEEE Transactions on Automatic Control 36(6):668-682, 1991.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Assign input to appropriate structs.
[sys, opts, ioformat] = ml_lti2struct(varargin{:});

if ioformat == 0
    nargoutchk(0, 5);
else
    nargoutchk(0, 2);
end

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'careopts')
    assert(isa(opts.careopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.careopts has to be a struct!');
else
    opts.careopts = struct();
end

if ml_field_set_to_value(opts, 'Gamma')
    ml_assert_posscalar(opts.Gamma, 'opts.Gamma');
else
    opts.Gamma = Inf;
end

if ml_field_set_to_value(opts, 'Method')
    if strcmpi(opts.Method, 'bfsr')
        mor_method = @ml_bfsr;
    elseif strcmpi(opts.Method, 'sr')
        mor_method = @ml_sr;
    else
        error('MORLAB:data', ...
            'The given computation method is not implemented!');
    end
else
    mor_method = @ml_sr;
end

if ml_field_set_to_value(opts, 'OrderComputation')
    if strcmpi(opts.OrderComputation, 'order')
        tselect = 0;
    elseif strcmpi(opts.OrderComputation, 'tolerance')
        tselect = 5;
    else
        error('MORLAB:data', ...
            'The desired order computation method is not implemented!');
    end
else
    tselect = 5;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARTIAL STABILIZATION AND SOLVE MATRIX EQUATIONS.                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Disturbance of scaling parameter.
opts.Gamma = 1.0001 * max(1.0, opts.Gamma);
gam        = sqrt(1 - 1 / opts.Gamma^2);

% Computation of matrices for Riccati equations.
Rb = eye(size(sys.C, 1)) + sys.D * sys.D';
Lb = chol(Rb);
Rc = eye(size(sys.B, 2)) + sys.D' * sys.D;
Lc = chol(Rc);

Ab = sys.A - sys.B * sys.D' * (Rb \ sys.C);
Ac = sys.A - (sys.B / Rc) * sys.D' * sys.C;
B  = sys.B / Lc;
C  = Lb' \ sys.C;

% Partial stabilization.
[K1, infoPARTSTAB_C] = ml_partstab(Ab', gam * C', opts);
[K2, infoPARTSTAB_O] = ml_partstab(Ac, gam * B, opts);

% Computation of the full-rank factors of the system Gramians.
opts.careopts.K0 = K1;
[R, infoCARE_C]  = ml_care_nwt_fac(Ab', gam * C', B', [], opts.careopts);

opts.careopts.K0 = K2;
[L, infoCARE_O]  = ml_care_nwt_fac(Ac, gam * B, C, [], opts.careopts);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COMPUTE REDUCED-ORDER MODEL.                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[sys, hsv] = mor_method(sys, R', L, tselect, opts);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN OUTPUT.                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Construction of the reduced-order model.
switch ioformat
    case 0
        varargout{1} = sys.A;
        varargout{2} = sys.B;
        varargout{3} = sys.C;
        varargout{4} = sys.D;
        
    case 1
        varargout{1} = struct(...
            'A', sys.A, ...
            'B', sys.B, ...
            'C', sys.C, ...
            'D', sys.D);
        
    case 2
        varargout{1} = ss(...
            sys.A, ...
            sys.B, ...
            sys.C, ...
            sys.D);
end

% Assign output information.
n  = size(sys.A, 1);

info = struct(...
    'AbsErrBound'   , ...
    2 * sum(hsv(n+1:end) ./ (1 + (gam * hsv(n+1:end)).^2)), ...
    'Gamma'         , opts.Gamma, ...
    'Hsv'           , hsv, ...
    'infoCARE_C'    , infoCARE_C, ...
    'infoCARE_O'    , infoCARE_O, ...
    'infoPARTSTAB_C', infoPARTSTAB_C, ...
    'infoPARTSTAB_O', infoPARTSTAB_O, ...
    'N'             , n);

if ioformat == 0
    varargout{5} = info;
else
    varargout{2} = info;
end
