function [varargout] = ml_hna_dss(varargin)
%ML_HNA_DSS Hankel-norm approximation for descriptor systems.
%
% DESCRIPTION:
%   This function computes the generalized Hankel-norm approximation of a
%   descriptor system
%
%       E*x'(t) = A*x(t) + B*u(t),                                      (1)
%          y(t) = C*x(t) + D*u(t).                                      (2)
%
%   Therefore, first a balanced realization is computed by using the
%   generalized balanced truncation square-root method with an appropriate
%   tolerance for the minimal realization of the given system. Then the
%   strictly proper part of the system is transformed using the formulas
%   for all-pass systems. As result, a reduced-order system of the form
%
%       Er*x'(t) = Ar*x(t) + Br*u(t),                                   (3)
%           y(t) = Cr*x(t) + Dr*u(t)                                    (4)
%
%   is computed, such that for the original transfer function G and the
%   reduced-order transfer function Gr with an r-th order strictly proper
%   part it holds
%
%       ||G - Gr||_{H}       = Hsvp(r+1),
%       ||G - Gr||_{\infty} <= 2 * (Hsvp(r+1) + ... + Hsvp(n)),
%
%   with Hsvp, a vector containing the proper Hankel singular values of the
%   system.
%
%   Note: For unstable systems, an additional additive decomposition into
%         the stable and anti-stable parts is performed and then only the
%         stable part will be reduced. That does not change the error
%         formulas.
%
% 1) If the descriptor system is given by matrices:
% SIGNATURE:
%   [Ar, Br, Cr, Dr, Er, info] = ML_HNA_DSS(A, B, C, D, E)
%   [Ar, Br, Cr, Dr, Er, info] = ML_HNA_DSS(A, B, C, D, E, opts)
%
% INPUT:
%   A    - matrix from (1) with dimensions n x n
%   B    - matrix from (1) with dimensions n x m
%   C    - matrix from (2) with dimensions p x n
%   D    - matrix from (2) with dimensions p x m
%   E    - matrix from (1) with dimensions n x n
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | DecompTol       | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest singular value of E to determine the      |
%   |                 | smallest non-quasi-zero singular value of E       |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | gdlyapopts      | structure, containing the optional parameters for |
%   |                 | the computation of the generalized discrete-time  |
%   |                 | Lyapunov equations, with                          |
%   |                 |   AbsTol  - tolerance for the absolute gain       |
%   |                 |             (default 0)                           |
%   |                 |   Index   - index of nilpotency, Inf if unknown   |
%   |                 |             (default Inf)                         |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative gain       |
%   |                 |             (default 1.0e+01*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | hankeldiskopts  | structure, containing the optional parameters for |
%   |                 | the disk function used for the decomposition      |
%   |                 | after the transformation to an all-pass system    |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | ImproperTrunc   | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest proper Hankel singular value of the       |
%   |                 | system to truncate the improper part              |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | Index           | nonnegative integer, index of the descriptor      |
%   |                 | system used to set an upper bound on the size of  |
%   |                 | the reduced improper part, Inf if unknown         |
%   |                 | (default Inf)                                     |
%   |-----------------|---------------------------------------------------|
%   | infdecopts      | structure, containing the optional parameters for |
%   |                 | the decomposition of the finite and infinite parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the infinite       |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | lyapdlopts      | structure, containing the optional parameters for |
%   |                 | the computation of the generalized continuous-time|
%   |                 | Lyapunov equations, with                          |
%   |                 |   AbsTol  - tolerance for the absolute error      |
%   |                 |             (default 0)                           |
%   |                 |   CompTol - tolerance for column and row          |
%   |                 |             compression                           |
%   |                 |             (default log(n)*eps)                  |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative error      |
%   |                 |             (default 1.0e+01*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | MinRelTol       | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest characteristic value to determine a       |
%   |                 | minimal realization                               |
%   |                 | (default log(length(Hsv))*eps)                    |
%   |-----------------|---------------------------------------------------|
%   | Order           | positive integer, order of the resulting          |
%   |                 | reduced-order model chosen by the user if         |
%   |                 | 'order' is set for OrderComputation               |
%   |                 | (default min(10,length(Hsvp)) + Nu + Ni)          |
%   |-----------------|---------------------------------------------------|
%   | OrderComputation| character array, determining the method for the   |
%   |                 | computation of the size of the reduced-order model|
%   |                 |  'order'     - take explicit order                |
%   |                 |  'tolerance' - using absolute error bound         |
%   |                 | (default 'tolerance')                             |
%   |-----------------|---------------------------------------------------|
%   | stabdecopts     | structure, containing the optional parameters for |
%   |                 | the decomposition of the stable and unstable parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the anti-stable    |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | Tolerance       | nonnegative scalar, tolerance used for the        |
%   |                 | computation of the size of the reduced-order model|
%   |                 | by an absolute error bound if 'tolerance' is set  |
%   |                 | for OrderComputation                              |
%   |                 | (default 1.0e-02)                                 |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   Ar   - matrix of (3) with dimensions r x r
%   Br   - matrix of (3) with dimensions r x m
%   Cr   - matrix of (4) with dimensions p x r
%   Dr   - matrix of (4) with dimensions p x m
%   Er   - matrix of (3) with dimensions r x r
%   info - structure, containing the following information:
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | AbsErrBound     | computed error bound for the absolute error of the|
%   |                 | reduced-order model in H-infinity norm            |
%   |-----------------|---------------------------------------------------|
%   | Hsvi            | a vector, containing the computed Hankel singular |
%   |                 | values of the improper part of the system         |
%   |-----------------|---------------------------------------------------|
%   | Hsvp            | a vector, containing the computed Hankel singular |
%   |                 | values of the proper part of the system           |
%   |-----------------|---------------------------------------------------|
%   | infoGADTF       | structure, containing information about the       |
%   |                 | additive decomposition of the system into its     |
%   |                 | infinite, finite stable and finite anti-stable    |
%   |                 | parts, with                                       |
%   |                 |   infoINFDISK  - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  infinite eigenvalues             |
%   |                 |   infoSTABDISK - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  anti-stable eigenvalues          |
%   |                 |   Ninf         - number of infinite eigenvalues   |
%   |                 |   Ns           - number of stable eigenvalues     |
%   |                 |   Nu           - number of anti-stable eigenvalues|
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_C    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper controllability Gramian, with    |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_O    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper observability Gramian, with      |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoHADISK      | structure, containing information about the       |
%   |                 | disk function method, with                        |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoLYAPDL      | structure, containing information about the       |
%   |                 | continuous-time dual Lyapunov equations solver,   |
%   |                 | with                                              |
%   |                 |   AbsErr         - absolute error of iteration    |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of iteration    |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | Ni              | Dimension of the improper part in the reduced-    |
%   |                 | order model                                       |
%   |-----------------|---------------------------------------------------|
%   | Np              | Dimension of the proper part in the reduced-order |
%   |                 | model                                             |
%   |-----------------|---------------------------------------------------|
%   | Nu              | Dimension of the unstable part in the reduced-    |
%   |                 | order model                                       |
%   |-----------------|---------------------------------------------------|
%   | Sigma           | Chosen proper Hankel singular value, exact        |
%   |                 | approximation error in the Hankel-norm            |
%   |-----------------|---------------------------------------------------|
%
% 2) If the descriptor system is given in a struct:
% SIGNATURE:
%   [rom, info] = ML_HNA_DSS(sys)
%   [rom, info] = ML_HNA_DSS(sys, opts)
%
% INPUT:
%   sys  - structure, containing the descriptor system's matrices
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (1) with dimensions n x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (2) with dimensions p x n             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (2) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - structure, containing the reduced-order descriptor system
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (3) with dimensions r x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (4) with dimensions p x r             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (4) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing information (see above 1))
%
% 3) If the descriptor system is given as state-space object:
% SIGNATURE:
%   [rom, info] = ML_HNA_DSS(sys)
%   [rom, info] = ML_HNA_DSS(sys, opts)
%
% INPUT:
%   sys  - state-space object, containing the descriptor system's matrices
%          (see above 2))
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - state-space object, containing the reduced-order descriptor
%          system (see above 2))
%   info - structure, containing information (see above 1))
%
%
% REFERENCE:
%   S. WERNER. Hankel-norm approximation of descriptor systems. Master's
%   thesis, Otto von Guericke University, Magdeburg, Germany, 2016.
%   http://nbn-resolving.de/urn:nbn:de:gbv:ma9:1-8845.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Assign input to appropriate structs.
[sys, opts, ioformat] = ml_lti2struct_dss(varargin{:});

if ioformat == 0
    nargoutchk(0, 6);
else
    nargoutchk(0, 2);
end

% Fill the D matrix.
if isempty(sys.D)
    sys.D = zeros(size(sys.C, 1), size(sys.B, 2));
end

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'gdlyapopts')
    assert(isa(opts.gdlyapopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.gdlyapopts has to be a struct!');
else
    opts.gdlyapopts = struct();
end

if ml_field_set_to_value(opts, 'hankeldecopts')
    assert(isa(opts.hankeldecopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.hankeldecopts has to be a struct!');
else
    opts.hankeldecopts = struct();
end

if ml_field_set_to_value(opts, 'lyapdlopts')
    assert(isa(opts.lyapdlopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.lyapdlopts has to be a struct!');
else
    opts.lyapdlopts = struct();
end

if ml_field_set_to_value(opts, 'OrderComputation')
    if strcmpi(opts.OrderComputation, 'order')
        tselect = 0;
    elseif strcmpi(opts.OrderComputation, 'tolerance')
        tselect = 2;
    else
        error('MORLAB:data', ...
            'The desired order computation method is not implemented!');
    end
else
    tselect = 2;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION OF TRANSFER FUNCTION.                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Decomposition of the system into unstable, finite and infinite parts.
[sys, infoGADTF] = ml_adtf_dss(sys, opts);

ninf = size(sys.Ainf, 1);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COMPUTE MINIMAL, BALANCED REALIZATION.                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation of the full-rank factors of the proper Gramians.
if size(sys.A, 1) > 0
    [R, L, infoLYAPDL] = ml_lyapdl_sgn_fac(sys.A, sys.B, sys.C, sys.E, ...
        opts.lyapdlopts);
else
    [R, L, infoLYAPDL] = deal([]);
end

% Computation of the full-rank factors of the improper Gramians.
if ninf > 0
    [Rinf, infoGDLYAP_C] = ml_gdlyap_smith_fac(sys.Ainf, sys.Binf, ...
        sys.Einf, opts.gdlyapopts);
    [Linf, infoGDLYAP_O] = ml_gdlyap_smith_fac(sys.Ainf', sys.Cinf', ...
        sys.Einf', opts.gdlyapopts);
else
    [Rinf, Linf, infoGDLYAP_C, infoGDLYAP_O] = deal([]);
end

[sys, hsvp, hsvi] = ml_sr_dss(sys, R, Rinf, L, Linf, 1, opts);

nf = size(sys.A, 1);
nh = ml_order(hsvp, size(sys.Au, 1), size(sys.Ainf, 1), tselect, opts);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TRANSFORMATION TO REDUCED-ORDER MODEL.                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nh < nf
    sigma = hsvp(nh + 1);
    
    % Get multiplicity of the chosen Hankel singular value.
    k = 1;
    
    while k < nf - nh
        if abs(sigma - hsvp(nh + 1 + k)) > sigma * log(nf) * sqrt(eps)
            break;
        end
        
        k = k + 1;
    end
    
    % Permutation of the system.
    if nh + k < nf
        p = [1:nh, nh+k+1:nf, nh+1:nh+k];
        
        sys.A = sys.A(p,p);
        sys.B = sys.B(p, :);
        sys.C = sys.C(: , p);
        hsvh  = hsvp(p);
    else
        hsvh = hsvp(1:nf);
    end
    
    % Transformation to an all-pass error system.
    nk = nf - k;
    
    U = sigma * pinv(sys.C(: , nk+1:nf)') * sys.B(nk+1:nf, :);
    S = sparse(1:nk, 1:nk, hsvh(1:nk));
    A = sys.A(1:nk, 1:nk);
    B = sys.B(1:nk, :);
    C = sys.C(: , 1:nk);
    
    rom.A = sigma^2 * A' + S * A * S + C' * U * B';
    rom.B = S * B - C' * U;
    rom.C = C * S - U * B';
    rom.D = sys.D + U;
    rom.E = diag(hsvh(1:nk).^2 - sigma^2);
    
    % Additive decomposition.
    hankelopts                       = struct();
    hankelopts.infdecopts.Dimension  = 0;
    hankelopts.stabdecopts           = opts.hankeldecopts;
    hankelopts.stabdecopts.Dimension = nk - nh;
    
    [rom, infoHANKEL] = ml_adtf_dss(rom, hankelopts);
else
    sigma = 0;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN OUTPUT.                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Construction of the reduced-order model.
switch ioformat
    case 0
        varargout{1} = blkdiag(rom.A, sys.Au, sys.Ainf);
        varargout{2} = [rom.B; sys.Bu; sys.Binf];
        varargout{3} = [rom.C, sys.Cu, sys.Cinf];
        varargout{4} = rom.D;
        varargout{5} = blkdiag(rom.E, sys.Eu, sys.Einf);
        
    case 1
        varargout{1} = struct(...
            'A', blkdiag(rom.A, sys.Au, sys.Ainf), ...
            'B', [rom.B; sys.Bu; sys.Binf], ...
            'C', [rom.C, sys.Cu, sys.Cinf], ...
            'D', rom.D, ...
            'E', blkdiag(rom.E, sys.Eu, sys.Einf));
        
    case 2
        varargout{1} = dss(...
            blkdiag(rom.A, sys.Au, sys.Ainf), ...
            [rom.B; sys.Bu; sys.Binf], ...
            [rom.C, sys.Cu, sys.Cinf], ...
            rom.D, ...
            blkdiag(rom.E, sys.Eu, sys.Einf));
end

% Assign output information.
info = struct(...
    'AbsErrBound' , 2 * sum(hsvp(nh+1:end)), ...
    'Hsvi'        , hsvi, ...
    'Hsvp'        , hsvp, ...
    'infoGADTF'   , infoGADTF, ...
    'infoGDLYAP_C', infoGDLYAP_C, ...
    'infoGDLYAP_O', infoGDLYAP_O, ...
    'infoHADISK'  , infoHANKEL.infoSTABDISK, ...
    'infoLYAPDL'  , infoLYAPDL, ...
    'Ni'          , size(sys.Ainf, 1), ...
    'Np'          , nh, ...
    'Nu'          , size(sys.Au, 1), ...
    'Sigma'       , sigma);

if ioformat == 0
    varargout{6} = info;
else
    varargout{2} = info;
end
