function [varargout] = ml_lqgbt_dss(varargin)
%ML_LQGBT_DSS LQG balanced truncation for descriptor systems.
%
% DESCRIPTION:
%   This function computes the generalized linear-quadratic Gaussian
%   balanced truncation for a descriptor system of the form
%
%       E*x'(t) = A*x(t) + B*u(t),                                      (1)
%          y(t) = C*x(t) + D*u(t).                                      (2)
%
%   Therefore, first an additive decomposition of the system is performed
%   using the matrix disk function, such that
%
%            [ Ei  0 ]       [ Ai  0 ]                        [ Ci ]
%       E2 = [       ], A2 = [       ], B2 = [ Bi, Bp ], C2 = [    ],
%            [ 0  Ep ]       [ 0  Ap ]                        [ Cp ]
%
%   with (Ei, Ai, Bi, Ci, D) belonging to the polynomial part and
%   (Ep, Ap, Bp, Cp, 0) belonging to the strictly proper part.
%   Now, the two generalized continuous-time Riccati equations
%
%       Ap*Pp*Ep' + Ep*Pp*Ap' + Bp*Bp'
%           - (Ep*Pp*Cp' + B*M') * inv(Rb) * (Ep*Pp*Cp' + B*M')' = 0,
%       Ap'*Qp*Ep + Ep'*Qp*Ap + Cp'*Cp
%           - (Bp'*Qp*Ep + M'*C)' * inv(Rc) * (Bp'*Qp*Ep + M'*C) = 0
%
%   are solved for the reduction of the strictly proper part, with
%
%       Rb = I + M*M',
%       Rc = I + M'*M,
%
%   where M = D - Ci * inv(Ai) * Bi. Also, the two generalized
%   discrete-time Lyapunov equations
%
%       Ai*Pi*Ai' - Ei*Pi*Ei' - Bi*Bi' = 0,
%       Ai'*Qi*Ai - Ei'*Qi*Ei - Ci'*Ci = 0
%
%   are solved for the reduction of the polynomial part. As result, a
%   reduced-order system of the form
%
%       Er*x'(t) = Ar*x(t) + Br*u(t),                                   (3)
%           y(t) = Cr*x(t) + Dr*u(t)                                    (4)
%
%   is computed, such that for the original transfer function G with the
%   (right) coprime factorization G = N*inv(M) and the reduced-order
%   transfer function Gr with an r-th order strictly proper part and the
%   (right) coprime factorization Gr = Nr*inv(Mr) it holds
%
%       ||[N; M] - [Nr; Mr]||_{\infty} <= 2 * (Hsvp(r+1)/sqrt(1
%           + Hsvp(r+1)^2) + ... + Hsvp(n)/sqrt(1 + Hsvp(n)^2)),
%
%   with Hsvp the vector containing the characteristic LQG singular values
%   of the system.
%
% 1) If the descriptor system is given by matrices:
% SIGNATURE:
%   [Ar, Br, Cr, Dr, Er, info] = ML_LQGBT_DSS(A, B, C, D, E)
%   [Ar, Br, Cr, Dr, Er, info] = ML_LQGBT_DSS(A, B, C, D, E, opts)
%
% INPUT:
%   A    - matrix from (1) with dimensions n x n
%   B    - matrix from (1) with dimensions n x m
%   C    - matrix from (2) with dimensions p x n
%   D    - matrix from (2) with dimensions p x m
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | Beta            | nonnegative scalar, used as shift of the          |
%   |                 | in Bass' algorithm for better conditioning if     |
%   |                 | StabMethod == 'lyap' is chosen                    |
%   |                 | (default 0.1)                                     |
%   |-----------------|---------------------------------------------------|
%   | careopts        | structure, containing the optional parameters for |
%   |                 | the computation of the continuous-time algebraic  |
%   |                 | Riccati equations, with                           |
%   |                 |   AbsTol   - tolerance for the absolute error     |
%   |                 |              (default 0)                          |
%   |                 |   Info     - disable/enable display of verbose    |
%   |                 |              status information during iteration  |
%   |                 |              (default 0)                          |
%   |                 |   lyapopts - structure, containing the optional   |
%   |                 |              parameters for the Lyapunov solver   |
%   |                 |              used in each iteration step          |
%   |                 |     * AbsTol  - tolerance for the absolute error  |
%   |                 |                 (default 0)                       |
%   |                 |     * CompTol - tolerance for column compression  |
%   |                 |                 (default log(n)*eps)              |
%   |                 |     * Info    - disable/enable display of verbose |
%   |                 |                 status information                |
%   |                 |                 (default 0)                       |
%   |                 |     * MaxIter - maximum number of iteration steps |
%   |                 |                 (default 100)                     |
%   |                 |     * RelTol  - tolerance for the relative error  |
%   |                 |                 (default 1.0e+01*n*eps)           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative error      |
%   |                 |             (default 1.0e+02*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | DecompTol       | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest singular value of E to determine the      |
%   |                 | smallest non-quasi-zero singular value of E       |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | gdlyapopts      | structure, containing the optional parameters for |
%   |                 | the computation of the generalized discrete-time  |
%   |                 | Lyapunov equations, with                          |
%   |                 |   AbsTol  - tolerance for the absolute gain       |
%   |                 |             (default 0)                           |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   Info    - display information during iteration  |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative gain       |
%   |                 |             (default 1.0e+01*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | ImproperTrunc   | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest proper Hankel singular value of the       |
%   |                 | system to truncate the improper part              |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | Index           | nonnegative integer, index of the descriptor      |
%   |                 | system used to set an upper bound on the size of  |
%   |                 | the reduced improper part, Inf if unknown         |
%   |                 | (default Inf)                                     |
%   |-----------------|---------------------------------------------------|
%   | infdecopts      | structure, containing the optional parameters for |
%   |                 | the decomposition of the finite and infinite parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the infinite       |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | Method          | character array, determining algorithm for the    |
%   |                 | computation of the reduced-order model            |
%   |                 |  'sr'   - square-root method                      |
%   |                 |  'bfsr' - balancing-free square-root method       |
%   |                 | (default 'sr')                                    |
%   |-----------------|---------------------------------------------------|
%   | Order           | positive integer, order of the resulting          |
%   |                 | reduced-order model chosen by the user if         |
%   |                 | 'order' is set for OrderComputation               |
%   |                 | (default min(10,length(Hsv)) + Nu + Ni)           |
%   |-----------------|---------------------------------------------------|
%   | OrderComputation| character array, determining the method for the   |
%   |                 | computation of the size of the reduced-order model|
%   |                 |  'order'     - take explicit order                |
%   |                 |  'tolerance' - using absolute error bound         |
%   |                 | (default 'tolerance')                             |
%   |-----------------|---------------------------------------------------|
%   | stabdecopts     | structure, containing the optional parameters for |
%   |                 | the decomposition of the stable and unstable parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the anti-stable    |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | StabMethod      | character array, determining the method of        |
%   |                 | stabilization of the system                       |
%   |                 |   'abe'  - partial stabilization with the         |
%   |                 |            algebraic Bernoulli equation           |
%   |                 |   'lyap' - partial stabilization with Bass'       |
%   |                 |            algorithm                              |
%   |                 | (default 'abe')                                   |
%   |-----------------|---------------------------------------------------|
%   | stabmethodopts  | structure, containing the optional parameters for |
%   |                 | the sign function based Lyapunov or Bernoulli     |
%   |                 | equation solver used for the stabilization, with  |
%   |                 |   AbsTol    - tolerance for the absolute error    |
%   |                 |               (default 0)                         |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RelTol    - tolerance for the relative error    |
%   |                 |               (default 1.0e+01*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | Tolerance       | nonnegative scalar, tolerance used for the        |
%   |                 | computation of the size of the reduced-order model|
%   |                 | by an absolute error bound if 'tolerance' is set  |
%   |                 | for OrderComputation                              |
%   |                 | (default 1.0e-02)                                 |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   Ar   - matrix of (3) with dimensions r x r
%   Br   - matrix of (3) with dimensions r x m
%   Cr   - matrix of (4) with dimensions p x r
%   Dr   - matrix of (4) with dimensions p x m
%   Er   - matrix of (3) with dimensions r x r
%   info - structure, containing the following information:
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | AbsErrBound     | computed error bound for the absolute error of the|
%   |                 | (right) coprime factorization in H-infinity norm  |
%   |-----------------|---------------------------------------------------|
%   | Hsvi            | a vector, containing the computed Hankel singular |
%   |                 | values of the improper part of the system         |
%   |-----------------|---------------------------------------------------|
%   | Hsvp            | a vector, containing the computed characteristic  |
%   |                 | LQG singular values of the proper part of the     |
%   |                 | system                                            |
%   |-----------------|---------------------------------------------------|
%   | infoCARE_C      | structure, containing information about the       |
%   |                 | continuous-time algebraic Riccati equation solver |
%   |                 | for the for the controllability Gramian, with     |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | infoCARE_O      | structure, containing information about the       |
%   |                 | continuous-time algebraic Riccati equation solver |
%   |                 | for the for the observability Gramian, with       |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | infoGADTF       | structure, containing information about the       |
%   |                 | additive decomposition of the system into its     |
%   |                 | infinite, finite stable and finite anti-stable    |
%   |                 | parts, with                                       |
%   |                 |   infoINFDISK  - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  infinite eigenvalues             |
%   |                 |   infoSTABDISK - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  anti-stable eigenvalues          |
%   |                 |   Ninf         - number of infinite eigenvalues   |
%   |                 |   Ns           - number of stable eigenvalues     |
%   |                 |   Nu           - number of anti-stable eigenvalues|
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_C    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper controllability Gramian, with    |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_O    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper observability Gramian, with      |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoPARTSTAB_C  | structure, containing information about the       |
%   |                 | partial stabilization used for the controllability|
%   |                 | Riccati equation                                  |
%   |                 |   infoSTABMETH   - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    stabilization                  |
%   |                 |   infoSTABMETH2  - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    a second stabilization         |
%   |                 |   infoSTABSIGNM  - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    separation of anti-stable      |
%   |                 |                    eigenvalues                    |
%   |                 |   infoSTABSIGNM2 - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    a second separation of anti-   |
%   |                 |                    stable eigenvalues             |
%   |                 |   Method         - character array with shortcut  |
%   |                 |                    of the used method             |
%   |                 |   Ns             - identified stable eigenvalues  |
%   |                 |   Nu             - identified anti-stable         |
%   |                 |                    eigenvalues                    |
%   |-----------------|---------------------------------------------------|
%   | infoPARTSTAB_O  | structure, containing information about the       |
%   |                 | partial stabilization used for the observability  |
%   |                 | Riccati equation                                  |
%   |                 |   infoSTABMETH   - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    stabilization                  |
%   |                 |   infoSTABMETH2  - structure with information     |
%   |                 |                    about sign function solver for |
%   |                 |                    a second stabilization         |
%   |                 |   infoSTABSIGNM  - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    separation of anti-stable      |
%   |                 |                    eigenvalues                    |
%   |                 |   infoSTABSIGNM2 - structure with information     |
%   |                 |                    about sign function method for |
%   |                 |                    a second separation of anti-   |
%   |                 |                    stable eigenvalues             |
%   |                 |   Method         - character array with shortcut  |
%   |                 |                    of the used method             |
%   |                 |   Ns             - identified stable eigenvalues  |
%   |                 |   Nu             - identified anti-stable         |
%   |                 |                    eigenvalues                    |
%   |-----------------|---------------------------------------------------|
%   | Nf              | Dimension of the finite reduced part in the       |
%   |                 | reduced-order model                               |
%   |-----------------|---------------------------------------------------|
%   | Ni              | Dimension of the improper part in the reduced-    |
%   |                 | order model                                       |
%   |-----------------|---------------------------------------------------|
%
% 2) If the descriptor system is given in a struct:
% SIGNATURE:
%   [rom, info] = ML_LQGBT_DSS(sys)
%   [rom, info] = ML_LQGBT_DSS(sys, opts)
%
% INPUT:
%   sys  - structure, containing the descriptor system's matrices
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (1) with dimensions n x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (2) with dimensions p x n             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (2) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - structure, containing the reduced-order descriptor system
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (3) with dimensions r x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (4) with dimensions p x r             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (4) with dimensions p x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing information (see above 1))
%
% 3) If the descriptor system is given as state-space object:
% SIGNATURE:
%   [rom, info] = ML_LQGBT_DSS(sys)
%   [rom, info] = ML_LQGBT_DSS(sys, opts)
%
% INPUT:
%   sys  - state-space object, containing the descriptor system's matrices
%          (see above 2))
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - state-space object, containing the reduced-order descriptor
%          system (see above 2))
%   info - structure, containing information (see above 1))
%
%
% REFERENCE:
%   P. BENNER, T. STYKEL. Model Order Reduction for
%   Differential-Algebraic Equations: A Survey. In: Ilchmann A.,
%   Reis T. (eds) Surveys in Differential-Algebraic Equations IV.
%   Differential-Algebraic Equations Forum. Springer, Cham, 2017.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Assign input to appropriate structs.
[sys, opts, ioformat] = ml_lti2struct_dss(varargin{:});

if ioformat == 0
    nargoutchk(0, 6);
else
    nargoutchk(0, 2);
end

% Fill the D matrix.
if isempty(sys.D)
    sys.D = zeros(size(sys.C, 1), size(sys.B, 2));
end

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'careopts')
    assert(isa(opts.careopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.careopts has to be a struct!');
else
    opts.careopts = struct();
end

if ml_field_set_to_value(opts, 'gdlyapopts')
    assert(isa(opts.gdlyapopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.gdlyapopts has to be a struct!');
else
    opts.gdlyapopts = struct();
end

if ml_field_set_to_value(opts, 'infdecopts')
    assert(isa(opts.infdecopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.infdecopts has to be a struct!');
else
    opts.infdecopts = struct();
end

if ml_field_set_to_value(opts, 'Method')
    if strcmpi(opts.Method, 'bfsr')
        mor_method = @ml_bfsr_dss;
    elseif strcmpi(opts.Method, 'sr')
        mor_method = @ml_sr_dss;
    else
        error('MORLAB:data', ...
            'The given computation method is not implemented!');
    end
else
    mor_method = @ml_sr_dss;
end

if ml_field_set_to_value(opts, 'OrderComputation')
    if strcmpi(opts.OrderComputation, 'order')
        tselect = 0;
    elseif strcmpi(opts.OrderComputation, 'tolerance')
        tselect = 3;
    else
        error('MORLAB:data', ...
            'The desired order computation method is not implemented!');
    end
else
    tselect = 3;
end

if ml_field_set_to_value(opts, 'stabdecopts')
    assert(isa(opts.stabdecopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.stabdecopts has to be a struct!');
else
    opts.stabdecopts = struct();
end

% Handling of anti-stable dimension for decomposition or stabilization.
if ml_field_set_to_value(opts.stabdecopts, 'Dimension')
    ml_assert_integer(opts.stabdecopts.Dimension, ...
        'opts.stabdecopts.Dimension');
else
    opts.stabdecopts.Dimension = -1;
end

nu                         = opts.stabdecopts.Dimension;
opts.stabdecopts.Dimension = 0;


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION OF TRANSFER FUNCTION.                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Decomposition of the system into finite and infinite parts.
[sys, infoGADTF] = ml_adtf_dss(sys, opts);

ninf = size(sys.Ainf, 1);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARTIAL STABILIZATION AND SOLVE MATRIX EQUATIONS.                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation of the constant polynomial part.
if ninf > 0
    M = sys.D - sys.Cinf * (sys.Ainf \ sys.Binf);
else
    M = sys.D;
end

if size(sys.A, 1) > 0
    % Computation of matrices for Riccati equations.
    Rb = eye(size(sys.C, 1)) + M * M';
    Lb = chol(Rb);
    Rc = eye(size(sys.B, 2)) + M' * M;
    Lc = chol(Rc);
    
    Ab = sys.A - sys.B * M' * (Rb \ sys.C);
    Ac = sys.A - (sys.B / Rc) * M' * sys.C;
    B  = sys.B / Lc;
    C  = Lb' \ sys.C;
    
    % Partial stabilization.
    opts.stabdecopts.Dimension = nu;
    opts.infdecopts.Dimension  = 0;
    
    [K1, infoPARTSTAB_C] = ml_partstab_dss(Ab', C', sys.E', opts);
    [K2, infoPARTSTAB_O] = ml_partstab_dss(Ac, B, sys.E, opts);
    
    % Computation of the full-rank factors of the proper Gramians.
    opts.careopts.K0 = K1;
    [R, infoCARE_C]  = ml_care_nwt_fac(Ab', C', B', sys.E', opts.careopts);
    
    opts.careopts.K0 = K2;
    [L, infoCARE_O]  = ml_care_nwt_fac(Ac, B, C, sys.E, opts.careopts);
else
    [R, L, infoPARTSTAB_C, infoPARTSTAB_O, infoCARE_C, infoCARE_O] = ...
        deal([]);
end

% Computation of the full-rank factors of the improper Gramians.
if ninf > 0
    [Rinf, infoGDLYAP_C] = ml_gdlyap_smith_fac(sys.Ainf, sys.Binf, ...
        sys.Einf, opts.gdlyapopts);
    [Linf, infoGDLYAP_O] = ml_gdlyap_smith_fac(sys.Ainf', sys.Cinf', ...
        sys.Einf', opts.gdlyapopts);
else
    [Rinf, Linf, infoGDLYAP_C, infoGDLYAP_O] = deal([]);
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COMPUTE REDUCED-ORDER MODEL.                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[sys, hsvp, hsvi] = mor_method(sys, R', Rinf, L, Linf, tselect, opts);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN OUTPUT.                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Construction of the reduced-order model.
switch ioformat
    case 0
        varargout{1} = blkdiag(sys.A, sys.Ainf);
        varargout{2} = [sys.B; sys.Binf];
        varargout{3} = [sys.C, sys.Cinf];
        varargout{4} = sys.D;
        varargout{5} = blkdiag(sys.E, sys.Einf);
        
    case 1
        varargout{1} = struct(...
            'A', blkdiag(sys.A, sys.Ainf), ...
            'B', [sys.B; sys.Binf], ...
            'C', [sys.C, sys.Cinf], ...
            'D', sys.D, ...
            'E', blkdiag(sys.E, sys.Einf));
        
    case 2
        varargout{1} = dss(...
            blkdiag(sys.A, sys.Ainf), ...
            [sys.B; sys.Binf], ...
            [sys.C, sys.Cinf], ...
            sys.D, ...
            blkdiag(sys.E, sys.Einf));
end

% Assign output information.
nf = size(sys.A, 1);

info = struct(...
    'AbsErrBound'   , ...
    2 * sum(hsvp(nf+1:end) ./ (1 + hsvp(nf+1:end).^2)), ...
    'Hsvi'          , hsvi, ...
    'Hsvp'          , hsvp, ...
    'infoCARE_C'    , infoCARE_C, ...
    'infoCARE_O'    , infoCARE_O, ...
    'infoGADTF'     , infoGADTF, ...
    'infoGDLYAP_C'  , infoGDLYAP_C, ...
    'infoGDLYAP_O'  , infoGDLYAP_O, ...
    'infoPARTSTAB_C', infoPARTSTAB_C, ...
    'infoPARTSTAB_O', infoPARTSTAB_O, ...
    'Nf'            , nf, ...
    'Ni'            , size(sys.Ainf, 1));

if ioformat == 0
    varargout{6} = info;
else
    varargout{2} = info;
end
