function [varargout] = ml_prbt_dss(varargin)
%ML_PRBT_DSS Positive-real balanced truncation for descriptor systems.
%
% DESCRIPTION:
%   This function computes the generalized positive-real balanced
%   truncation for a positive-real descriptor system of the form
%
%       E*x'(t) = A*x(t) + B*u(t),                                      (1)
%          y(t) = C*x(t) + D*u(t).                                      (2)
%
%   Therefore, first an additive decomposition of the system is performed
%   using the matrix disk function, such that
%
%            [ Ei  0 ]       [ Ai  0 ]                        [ Ci ]
%       E2 = [       ], A2 = [       ], B2 = [ Bi, Bp ], C2 = [    ],
%            [ 0  Ep ]       [ 0  Ap ]                        [ Cp ]
%
%   with (Ei, Ai, Bi, Ci, D) belonging to the polynomial part and
%   (Ep, Ap, Bp, Cp, 0) belonging to the strictly proper part.
%   Now, the two generalized positive-real Riccati equations
%
%       Ap*Pp*Ep' + Ep*Pp*Ap'
%           + (Ep*Pp*Cp' + Bp) * inv(R) * (Ep*Pp*Cp' + Bp)' = 0,
%       Ap'*Qp*Ep + Ep'*Qp*Ap + Cp'*Cp
%           + (Bp'*Qp*Ep + Cp)' * inv(R) * (Bp'*Qp*Ep + Cp) = 0
%
%   are solved for the reduction of the strictly proper part, with
%
%       R = M + M',
%
%   where M = D - Ci * inv(Ai) * Bi. Also, the two generalized
%   discrete-time Lyapunov equations
%
%       Ai*Pi*Ai' - Ei*Pi*Ei' - Bi*Bi' = 0,
%       Ai'*Qi*Ai - Ei'*Qi*Ei - Ci'*Ci = 0
%
%   are solved for the reduction of the polynomial part. As result, a
%   reduced-order system of the form
%
%       Er*x'(t) = Ar*x(t) + Br*u(t),                                   (3)
%           y(t) = Cr*x(t) + Dr*u(t)                                    (4)
%
%   is computed, such that for the original transfer function G and the
%   reduced-order transfer function Gr with an r-th order strictly proper
%   part it holds
%
%       ||inv(G + M') - inv(Gr + M')||_{\infty} <= 2 * ||R||_{2}^2
%           * (Hsvp(r+1) + ... + Hsvp(n)),
%
%   with Hsvp, a vector containing the proper characteristic positive-real
%   singular values of the system, and R^2 = inv(M + M').
%
%   Note: In case of a rank-deficient M + M' term, an epsilon
%         regularization is performed, which replaces the M during the
%         computations with an identity matrix scaled by a given epsilon.
%
% 1) If the descriptor system is given by matrices:
% SIGNATURE:
%   [Ar, Br, Cr, Dr, Er, info] = ML_PRBT_DSS(A, B, C, D, E)
%   [Ar, Br, Cr, Dr, Er, info] = ML_PRBT_DSS(A, B, C, D, E, opts)
%
% INPUT:
%   A    - matrix from (1) with dimensions n x n
%   B    - matrix from (1) with dimensions n x m
%   C    - matrix from (2) with dimensions m x n
%   D    - matrix from (2) with dimensions m x m
%   E    - matrix from (1) with dimensions n x n
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | DecompTol       | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest singular value of E to determine the      |
%   |                 | smallest non-quasi-zero singular value of E       |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | Epsilon         | positive scalar, used in the case of a            |
%   |                 | non-full-rank M + M' term for epsilon             |
%   |                 | regularization by multiplying with an identity    |
%   |                 | matrix of appropriate size                        |
%   |                 | (default 1.0e-03)                                 |
%   |-----------------|---------------------------------------------------|
%   | gdlyapopts      | structure, containing the optional parameters for |
%   |                 | the computation of the generalized discrete-time  |
%   |                 | Lyapunov equations, with                          |
%   |                 |   AbsTol  - tolerance for the absolute gain       |
%   |                 |             (default 0)                           |
%   |                 |   Index   - index of nilpotency, Inf if unknown   |
%   |                 |             (default Inf)                         |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative gain       |
%   |                 |             (default 1.0e+01*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | ImproperTrunc   | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest proper Hankel singular value of the       |
%   |                 | system to truncate the improper part              |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | Index           | nonnegative integer, index of the descriptor      |
%   |                 | system used to set an upper bound on the size of  |
%   |                 | the reduced improper part, Inf if unknown         |
%   |                 | (default Inf)                                     |
%   |-----------------|---------------------------------------------------|
%   | infdecopts      | structure, containing the optional parameters for |
%   |                 | the decomposition of the finite and infinite parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the infinite       |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | Method          | character array, determining algorithm for the    |
%   |                 | computation of the reduced-order model            |
%   |                 |  'sr'   - square root method                      |
%   |                 |  'bfsr' - balancing-free square root method       |
%   |                 | (default 'sr')                                    |
%   |-----------------|---------------------------------------------------|
%   | Order           | positive integer, order of the resulting          |
%   |                 | reduced-order model chosen by the user if         |
%   |                 | 'order' is set for OrderComputation               |
%   |                 | (default min(10,length(Hsvp)) + Ni)               |
%   |-----------------|---------------------------------------------------|
%   | OrderComputation| character array, determining the method for the   |
%   |                 | computation of the size of the reduced-order model|
%   |                 |  'order'     - take explicit order                |
%   |                 |  'tolerance' - using absolute error bound         |
%   |                 | (default 'tolerance')                             |
%   |-----------------|---------------------------------------------------|
%   | pcareopts       | structure, containing the optional parameters for |
%   |                 | the computation of the continuous-time algebraic  |
%   |                 | positive Riccati equation, with                   |
%   |                 |   AbsTol   - tolerance for the absolute error     |
%   |                 |              (default 0)                          |
%   |                 |   CompTol  - tolerance for column and row         |
%   |                 |              compression                          |
%   |                 |              (default log(n)*eps)                 |
%   |                 |   Info     - disable/enable display of verbose    |
%   |                 |              status information during iteration  |
%   |                 |              (default 0)                          |
%   |                 |   lyapopts - structure, containing the optional   |
%   |                 |              parameters for the Lyapunov solver   |
%   |                 |              used in each iteration step          |
%   |                 |     * AbsTol  - tolerance for the absolute error  |
%   |                 |                 (default 0)                       |
%   |                 |     * CompTol - tolerance for column compression  |
%   |                 |                 (default log(n)*eps)              |
%   |                 |     * Info    - disable/enable display of verbose |
%   |                 |                 status information                |
%   |                 |                 (default 0)                       |
%   |                 |     * MaxIter - maximum number of iteration steps |
%   |                 |                 (default 100)                     |
%   |                 |     * RelTol  - tolerance for the relative error  |
%   |                 |                 (default 1.0e+01*n*eps)           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative error      |
%   |                 |             (default 1.0e+02*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | Tolerance       | nonnegative scalar, tolerance used for the        |
%   |                 | computation of the size of the reduced-order model|
%   |                 | by an absolute error bound if 'tolerance' is set  |
%   |                 | for OrderComputation                              |
%   |                 | (default 1.0e-02)                                 |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   Ar   - matrix of (3) with dimensions r x r
%   Br   - matrix of (3) with dimensions r x m
%   Cr   - matrix of (4) with dimensions m x r
%   Dr   - matrix of (4) with dimensions m x m
%   Er   - matrix of (3) with dimensions r x r
%   info - structure, containing the following information:
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | Hsvi            | a vector, containing the computed Hankel singular |
%   |                 | values of the improper part of the system         |
%   |-----------------|---------------------------------------------------|
%   | Hsvp            | a vector, containing the computed characteristic  |
%   |                 | positive-real singular values of the proper part  |
%   |                 | of the system                                     |
%   |-----------------|---------------------------------------------------|
%   | infoGADTF       | structure, containing information about the       |
%   |                 | additive decomposition of the system into its     |
%   |                 | infinite and finite parts, with                   |
%   |                 |   infoINFDISK  - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  infinite eigenvalues             |
%   |                 |   infoSTABDISK - structure with information about |
%   |                 |                  the disk function method for the |
%   |                 |                  anti-stable eigenvalues          |
%   |                 |   Ninf         - number of infinite eigenvalues   |
%   |                 |   Ns           - number of stable eigenvalues     |
%   |                 |   Nu           - number of anti-stable eigenvalues|
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_C    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper controllability Gramian, with    |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoGDLYAP_O    | structure, containing information about the       |
%   |                 | generalized discrete-time Lyapunov equation solver|
%   |                 | for the improper observability Gramian, with      |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoPCARE_C     | structure, containing information about the       |
%   |                 | continuous-time algebraic positive Riccati        |
%   |                 | equation solver for the controllability Gramian,  |
%   |                 | with                                              |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | infoPCARE_O     | structure, containing information about the       |
%   |                 | continuous-time algebraic positive Riccati        |
%   |                 | equation solver for the observability Gramian,    |
%   |                 | with                                              |
%   |                 |   AbsErr         - absolute error of residual     |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of residual     |
%   |-----------------|---------------------------------------------------|
%   | InvAbsErrBound  | computed error bound for the absolute error of the|
%   |                 | inverse transfer functions in H-infinity norm     |
%   |-----------------|---------------------------------------------------|
%   | M               | matrix with dimensions m x m, polynomial part of  |
%   |                 | zeroth order and used in the error bound          |
%   |                 | (or term from the epsilon regularization)         |
%   |-----------------|---------------------------------------------------|
%   | Ni              | Dimension of the improper part in the reduced-    |
%   |                 | order model                                       |
%   |-----------------|---------------------------------------------------|
%   | Np              | Dimension of the proper part in the reduced-order |
%   |                 | model                                             |
%   |-----------------|---------------------------------------------------|
%
% 2) If the descriptor system is given in a struct:
% SIGNATURE:
%   [rom, info] = ML_PRBT_DSS(sys)
%   [rom, info] = ML_PRBT_DSS(sys, opts)
%
% INPUT:
%   sys  - structure, containing the descriptor system's matrices
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (1) with dimensions n x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (2) with dimensions m x n             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (2) with dimensions m x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (1) with dimensions n x n             |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - structure, containing the reduced-order descriptor system
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   |        A        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   |        B        | matrix from (3) with dimensions r x m             |
%   |-----------------|---------------------------------------------------|
%   |        C        | matrix from (4) with dimensions m x r             |
%   |-----------------|---------------------------------------------------|
%   |        D        | matrix from (4) with dimensions m x m             |
%   |-----------------|---------------------------------------------------|
%   |        E        | matrix from (3) with dimensions r x r             |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing information (see above 1))
%
% 3) If the descriptor system is given as state-space object:
% SIGNATURE:
%   [rom, info] = ML_PRBT_DSS(sys)
%   [rom, info] = ML_PRBT_DSS(sys, opts)
%
% INPUT:
%   sys  - state-space object, containing the descriptor system's matrices
%          (see above 2))
%   opts - structure, containing optional parameters (see above 1))
%
% OUTPUT:
%   rom  - state-space object, containing the reduced-order descriptor
%          system (see above 2))
%   info - structure, containing information (see above 1))
%
%
% REFERENCE:
%   T. REIS, T. STYKEL. Positive real and bounded real balancing for model
%   reduction of descriptor systems. International Journal of Control,
%   83:74-88, 2010.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Assign input to appropriate structs.
[sys, opts, ioformat] = ml_lti2struct_dss(varargin{:});

if ioformat == 0
    nargoutchk(0, 6);
else
    nargoutchk(0, 2);
end

m = size(sys.B, 2);

assert(size(sys.C, 1) == m, ...
    'MORLAB:data', ...
    ['The system has to be square.\nThe matrix C must have the ' ...
    'same number of rows as columns of matrix B.']);

% Fill the D matrix.
if isempty(sys.D)
    sys.D = zeros(m);
end

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'Epsilon')
    ml_assert_posscalar(opts.Epsilon, 'opts.Epsilon');
else
    opts.Epsilon = 1.0e-03;
end

if ml_field_set_to_value(opts, 'gdlyapopts')
    assert(isa(opts.gdlyapopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.gdlyapopts has to be a struct!');
else
    opts.gdlyapopts = struct();
end

if ml_field_set_to_value(opts, 'Method')
    if strcmpi(opts.Method, 'bfsr')
        mor_method = @ml_bfsr_dss;
    elseif strcmpi(opts.Method, 'sr')
        mor_method = @ml_sr_dss;
    else
        error('MORLAB:data', ...
            'The given computation method is not implemented!');
    end
else
    mor_method = @ml_sr_dss;
end

if ml_field_set_to_value(opts, 'OrderComputation')
    if strcmpi(opts.OrderComputation, 'order')
        tselect = 0;
    elseif strcmpi(opts.OrderComputation, 'tolerance')
        tselect = 2;
    else
        error('MORLAB:data', ...
            'The desired order computation method is not implemented!');
    end
else
    tselect = 2;
end

if ml_field_set_to_value(opts, 'pcareopts')
    assert(isa(opts.pcareopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.pcareopts has to be a struct!');
else
    opts.pcareopts = struct();
end

if ml_field_set_to_value(opts, 'stabdec')
    assert(isa(opts.stabdec, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.stabdec has to be a struct!');
else
    opts.stabdec = struct();
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION OF TRANSFER FUNCTION.                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Decomposition of the system into finite and infinite parts.
opts.stabdec.Dimension = 0;
[sys, infoGADTF]       = ml_adtf_dss(sys, opts);

ninf = size(sys.Ainf, 1);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SOLVE MATRIX EQUATIONS.                                                 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation of the constant polynomial part.
if ninf > 0
    M = sys.D - sys.Cinf * (sys.Ainf \ sys.Binf);
else
    M = sys.D;
end

if size(sys.A, 1) > 0
    % Epsilon regularization.
    if (rank(M + M') < m)  || (norm(M, 'fro') < 1.0e+02 * eps)
        M = opts.Epsilon * eye(m);
        J = 2 * opts.Epsilon * eye(m);
        D = sqrt(2 * opts.Epsilon) * eye(m);
    else
        J = M + M';
        D = chol(J);
    end
    
    % Computation of matrices for Riccati equations.
    A = sys.A - sys.B * (J \ sys.C);
    B = sys.B / D;
    C = D' \ sys.C;
    
    % Computation of the full-rank factors of the proper Gramians.
    [R, infoPCARE_C] = ml_pcare_nwt_fac(A', C', B', sys.E', ...
        opts.pcareopts);
    [L, infoPCARE_O] = ml_pcare_nwt_fac(A, B, C, sys.E, ...
        opts.pcareopts);
else
    J = 1;
    
    [R, L, infoPCARE_C, infoPCARE_O] = deal([]);
end

% Computation of the full-rank factors of the improper Gramians.
if ninf > 0
    [Rinf, infoGDLYAP_C] = ml_gdlyap_smith_fac(sys.Ainf, sys.Binf, ...
        sys.Einf, opts.gdlyapopts);
    [Linf, infoGDLYAP_O] = ml_gdlyap_smith_fac(sys.Ainf', sys.Cinf', ...
        sys.Einf', opts.gdlyapopts);
else
    [Rinf, Linf, infoGDLYAP_C, infoGDLYAP_O] = deal([]);
end

% Computation of the reduced-order model.
if ml_field_set_to_value(opts, 'Tolerance')
    ml_assert_nonnegscalar(opts.Tolerance, 'opts.Tolerance');
else
    opts.Tolerance = 1.0e-02;
end

opts.Tolerance = opts.Tolerance * min(svd(J));


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COMPUTE REDUCED-ORDER MODEL.                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[sys, hsvp, hsvi] = mor_method(sys, R', Rinf, L, Linf, tselect, opts);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN OUTPUT.                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Construction of the reduced-order model.
switch ioformat
    case 0
        varargout{1} = blkdiag(sys.A, sys.Ainf);
        varargout{2} = [sys.B; sys.Binf];
        varargout{3} = [sys.C, sys.Cinf];
        varargout{4} = sys.D;
        varargout{5} = blkdiag(sys.E, sys.Einf);
        
    case 1
        varargout{1} = struct(...
            'A', blkdiag(sys.A, sys.Ainf), ...
            'B', [sys.B; sys.Binf], ...
            'C', [sys.C, sys.Cinf], ...
            'D', sys.D, ...
            'E', blkdiag(sys.E, sys.Einf));
        
    case 2
        varargout{1} = dss(...
            blkdiag(sys.A, sys.Ainf), ...
            [sys.B; sys.Binf], ...
            [sys.C, sys.Cinf], ...
            sys.D, ...
            blkdiag(sys.E, sys.Einf));
end

% Assign output information.
nf = size(sys.A, 1);

info = struct(...
    'Hsvi'          , hsvi, ...
    'Hsvp'          , hsvp, ...
    'infoGADTF'     , infoGADTF, ...
    'infoGDLYAP_C'  , infoGDLYAP_C, ...
    'infoGDLYAP_O'  , infoGDLYAP_O, ...
    'infoPCARE_C'   , infoPCARE_C, ...
    'infoPCARE_O'   , infoPCARE_O, ...
    'InvAbsErrBound', 2 / min(svd(J)) * sum(hsvp(nf+1:end)), ...
    'M'             , M, ...
    'Ni'            , size(sys.Ainf, 1), ...
    'Np'            , nf);

if ioformat == 0
    varargout{6} = info;
else
    varargout{2} = info;
end
