function [sys, info] = ml_adtf(sys, opts)
%ML_ADTF Additive decomposition of standard system's transfer function.
%
% DESCRIPTION:
%   Consider a standard system of the form
%
%       x'(t) = A*x(t) + B*u(t),                                        (1)
%        y(t) = C*x(t).                                                 (2)
%
%   This function computes an additive decomposition of the corresponding
%   transfer function by a block diagonalization of the matrix A. The
%   system is transformed, such that
%
%               [ As     ]       [ Bs ]
%       z'(t) = [        ]z(t) + [    ]u(t),                            (3)
%               [     Au ]       [ Bu ]
%
%       y(t)  = [ Cs, Cu ]z(t),                                         (4)
%
%   where As contains the eigenvalues with negative real parts and Au all
%   the eigenvalues with positive real parts.
%
% SIGNATURE:
%   [sys, info] = ML_ADTF(sys)
%   [sys, info] = ML_ADTF(sys, opts)
%
% INPUT:
%   sys  - structure, containing the descriptor system in the form:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | A               | matrix with dimensions n x n in (1)               |
%   |-----------------|---------------------------------------------------|
%   | B               | matrix with dimensions n x m in (1)               |
%   |-----------------|---------------------------------------------------|
%   | C               | matrix with dimensions p x n in (2)               |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | stabsignmopts   | structure, containing the optional parameters for |
%   |                 | the matrix sign function used for the             |
%   |                 | decomposition into stable and anti-stable system  |
%   |                 | parts                                             |
%   |                 |   AbsTol  - tolerance for the absolute change     |
%   |                 |             (default 0)                           |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative change     |
%   |                 |             (default 1.0e+02*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | stabsylvopts    | structure, containing the optional parameters for |
%   |                 | the Sylvester equation solver used for the        |
%   |                 | decomposition into stable and anti-stable system  |
%   |                 | parts                                             |
%   |                 |   AbsTol  - tolerance for the absolute error      |
%   |                 |             (default 0)                           |
%   |                 |   Info    - disable/enable display of verbose     |
%   |                 |             status information during iteration   |
%   |                 |             (default 0)                           |
%   |                 |   MaxIter - maximum number of iteration steps     |
%   |                 |             (default 100)                         |
%   |                 |   RelTol  - tolerance for the relative error      |
%   |                 |             (default 1.0e+01*n*eps)               |
%   |-----------------|---------------------------------------------------|
%   | UnstabDim       | integer, dimension of the deflating anti-stable   |
%   |                 | subspace, negative if unknown                     |
%   |                 | (default -1)                                      |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   sys - structure, containing the transformed descriptor system:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | A               | matrix with dimensions ns x ns, see As in (3)     |
%   |-----------------|---------------------------------------------------|
%   | Au              | matrix with dimensions nu x nu in (3)             |
%   |-----------------|---------------------------------------------------|
%   | B               | matrix with dimensions ns x m, see Bs in (3)      |
%   |-----------------|---------------------------------------------------|
%   | Bu              | matrix with dimensions nu x m in (3)              |
%   |-----------------|---------------------------------------------------|
%   | C               | matrix with dimensions p x ns, see Cs in (4)      |
%   |-----------------|---------------------------------------------------|
%   | Cu              | matrix with dimensions p x nu in (4)              |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing the following information about the
%          algorithms used for the additive decomposition
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | infoSTABSIGNM   | structure, containing information about the       |
%   |                 | matrix sign function method, with                 |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoSTABSYLV    | structure, containing information about the       |
%   |                 | Sylvester equation solver, with                   |
%   |                 |   AbsErr         - absolute error of iteration    |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative error of iteration    |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | Ns              | Number of identified stable eigenvalues           |
%   |-----------------|---------------------------------------------------|
%   | Nu              | Number of identified anti-stable eigenvalues      |
%   |-----------------|---------------------------------------------------|
%
%
% REFERENCE:
%   B. KAGSTROM, P. VAN DOOREN. A generalized state-space approach for the
%   additive decomposition of a transfer matrix. Int. J. Numerical Linear
%   Algebra with Applications, 1(2):165-181, 1992.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

narginchk(1, 2);

if (nargin < 2) || isempty(opts)
    opts = struct();
end

% Check input standard system.
ml_assert_stdsys(sys);

n = size(sys.A, 1);

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'stabsignmopts')
    assert(isa(opts.stabsignmopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.stabsignmopts has to be a struct!');
else
    opts.stabsignmopts = struct();
end

if ml_field_set_to_value(opts, 'stabsylvopts')
    assert(isa(opts.stabsylvopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.stabsylvopts has to be a struct!');
else
    opts.stabsylvopts = struct();
end

if ml_field_set_to_value(opts, 'UnstabDim')
    ml_assert_integer(opts.UnstabDim, 'UnstabDim');
else
    opts.UnstabDim = -1;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION: ANTI-STABLE PART.                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation deflating subspaces using the matrix sign function.
if opts.UnstabDim == 0
    nu            = 0;
    infoSTABSIGNM = [];
else
    [Z, infoSTABSIGNM]  = ml_signm(sys.A, opts.stabsignmopts);
    
    if opts.UnstabDim > 0
        nu = opts.UnstabDim;
    else
        nu = (n + round(trace(Z))) / 2;
    end
end

% Block diagonalization of the system matrices.
if (nu > 0) && (nu < n)
    [Q, ~, ~] = qr(eye(n) + Z);
    
    sys.Au = Q(: , 1:nu)' * sys.A * Q(: , 1:nu);
    Wa     = Q(: , 1:nu)' * sys.A * Q(: , nu+1:end);
    sys.A  = Q(: , nu+1:end)' * sys.A * Q(: , nu+1:end);
    
    [Y, infoSTABSYLV] = ml_sylv_sgn(-sys.Au, sys.A, Wa, opts.stabsylvopts);
    
    sys.B  = Q' * sys.B;
    sys.Bu = sys.B(1:nu, :) - Y * sys.B(nu+1:end, :);
    sys.B  = sys.B(nu+1:end, :);
    
    sys.Cu = sys.C * Q;
    sys.C  = sys.Cu(: , nu+1:end) + sys.Cu(: , 1:nu) * Y;
    sys.Cu = sys.Cu(: , 1:nu);
elseif nu == n
    sys.Au = sys.A;
    sys.Bu = sys.B;
    sys.Cu = sys.C;
    
    [sys.A, sys.B, sys.C, infoSTABSYLV] = deal([]);
else
    [sys.Au, sys.Bu, sys.Cu, infoSTABSYLV] = deal([]);
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN INFORMATION.                                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Collection of all information.
info = struct(...
    'infoSTABSIGNM', infoSTABSIGNM, ...
    'infoSTABSYLV' , infoSTABSYLV, ...
    'Ns'           , n - nu, ...
    'Nu'           , nu);
