function [sys, info] = ml_adtf_dss(sys, opts)
%ML_ADTF_DSS Add. decomposition of descriptor system's transfer function.
%
% DESCRIPTION:
%   Consider a descriptor system of the form
%
%       E*x'(t) = A*x(t) + B*u(t),                                      (1)
%          y(t) = C*x(t).                                               (2)
%
%   This function computes an additive decomposition of the corresponding
%   transfer function by a block diagonalization of the matrix pencil
%   s*E - A. The descriptor system is transformed, such that
%
%       [ Ef         ]        [ Af           ]       [ Bf   ]
%       [    Einf    ]z'(t) = [     Ainf     ]z(t) + [ Binf ]u(t),      (3)
%       [         Eu ]        [           Au ]       [ Bu   ]
%
%                      y(t) = [ Cf, Cinf, Cu ]z(t),                     (4)
%
%   where s*Ef - Af contains the finite eigenvalues with negative real
%   parts, s*Einf - Ainf the infinite eigenvalues and s*Eu - Au the finite
%   eigenvalues with positive real parts.
%
% SIGNATURE:
%   [sys, info] = ML_ADTF_DSS(sys)
%   [sys, info] = ML_ADTF_DSS(sys, opts)
%
% INPUT:
%   sys  - structure, containing the descriptor system in the form:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | E               | matrix with dimensions n x n in (1)               |
%   |-----------------|---------------------------------------------------|
%   | A               | matrix with dimensions n x n in (1)               |
%   |-----------------|---------------------------------------------------|
%   | B               | matrix with dimensions n x m in (1)               |
%   |-----------------|---------------------------------------------------|
%   | C               | matrix with dimensions p x n in (2)               |
%   |-----------------|---------------------------------------------------|
%   opts - structure, containing the following optional entries:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | DecompTol       | nonnegative scalar, tolerance multiplied with the |
%   |                 | largest singular value of E to determine the      |
%   |                 | smallest non-quasi-zero singular value of E       |
%   |                 | (default log(n)*eps)                              |
%   |-----------------|---------------------------------------------------|
%   | infdecopts      | structure, containing the optional parameters for |
%   |                 | the decomposition of the finite and infinite parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the infinite       |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%   | stabdecopts     | structure, containing the optional parameters for |
%   |                 | the decomposition of the stable and unstable parts|
%   |                 | of the system using the disk function and subspace|
%   |                 | extraction method, with                           |
%   |                 |   AbsTol    - tolerance for the absolute change   |
%   |                 |               (default 0)                         |
%   |                 |   Dimension - dimension of the deflating subspace |
%   |                 |               corresponding to the anti-stable    |
%   |                 |               eigenvalues, negative if unknown    |
%   |                 |               (default -1)                        |
%   |                 |   Info      - disable/enable display of verbose   |
%   |                 |               status information during iteration |
%   |                 |               (default 0)                         |
%   |                 |   MaxIter   - maximum number of iteration steps   |
%   |                 |               (default 100)                       |
%   |                 |   RankTol   - tolerance used in numerical rank    |
%   |                 |               detection                           |
%   |                 |               (default  log(n)*eps)               |
%   |                 |   RelTol    - tolerance for the relative change   |
%   |                 |               (default 1.0e+02*n*eps)             |
%   |-----------------|---------------------------------------------------|
%
% OUTPUT:
%   sys - structure, containing the transformed descriptor system:
%   |-----------------|---------------------------------------------------|
%   |    PARAMETER    |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | E               | matrix with dimensions nf x nf, see Ef in (3)     |
%   |-----------------|---------------------------------------------------|
%   | Einf            | matrix with dimensions ninf x ninf in (3)         |
%   |-----------------|---------------------------------------------------|
%   | Eu              | matrix with dimensions nu x nu in (3)             |
%   |-----------------|---------------------------------------------------|
%   | A               | matrix with dimensions nf x nf, see Af in (3)     |
%   |-----------------|---------------------------------------------------|
%   | Ainf            | matrix with dimensions ninf x ninf in (3)         |
%   |-----------------|---------------------------------------------------|
%   | Au              | matrix with dimensions nu x nu in (3)             |
%   |-----------------|---------------------------------------------------|
%   | B               | matrix with dimensions nf x m, see Bf in (3)      |
%   |-----------------|---------------------------------------------------|
%   | Binf            | matrix with dimensions ninf x m in (3)            |
%   |-----------------|---------------------------------------------------|
%   | Bu              | matrix with dimensions nu x m in (3)              |
%   |-----------------|---------------------------------------------------|
%   | C               | matrix with dimensions p x nf, see Cf in (4)      |
%   |-----------------|---------------------------------------------------|
%   | Cinf            | matrix with dimensions p x ninf in (4)            |
%   |-----------------|---------------------------------------------------|
%   | Cu              | matrix with dimensions p x nu in (4)              |
%   |-----------------|---------------------------------------------------|
%   info - structure, containing the following information about the
%          additive decomposition of the descriptor system
%   |-----------------|---------------------------------------------------|
%   |      ENTRY      |                     MEANING                       |
%   |-----------------|---------------------------------------------------|
%   | infoINFDISK     | structure, containing information about the disk  |
%   |                 | function method used for the separation of the    |
%   |                 | infinite part, with                               |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | infoSTABDISK    | structure, containing information about the disk  |
%   |                 | function method used for the separation of the    |
%   |                 | unstable part, with                               |
%   |                 |   AbsErr         - absolute change of iteration   |
%   |                 |                    matrix                         |
%   |                 |   IterationSteps - number of iteration steps      |
%   |                 |   RelErr         - relative change of iteration   |
%   |                 |                    matrix                         |
%   |-----------------|---------------------------------------------------|
%   | Ninf            | Number of identified infinite eigenvalues         |
%   |-----------------|---------------------------------------------------|
%   | Ns              | Number of identified stable eigenvalues           |
%   |-----------------|---------------------------------------------------|
%   | Nu              | Number of identified anti-stable eigenvalues      |
%   |-----------------|---------------------------------------------------|
%
%
% REFERENCE:
%   B. KAGSTROM, P. VAN DOOREN. A generalized state-space approach for the
%   additive decomposition of a transfer matrix. Int. J. Numerical Linear
%   Algebra with Applications, 1(2):165-181, 1992.

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2006-2017 Peter Benner, Steffen W. R. Werner
%


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK INPUTS.                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

narginchk(1, 2);

if (nargin < 2) || isempty(opts)
    opts = struct();
end

% Check input descriptor system.
ml_assert_descsys(sys);

n = size(sys.A, 1);

% Check and assign optional parameters.
assert(isa(opts, 'struct'), ...
    'MORLAB:data', ...
    'The parameter opts has to be a struct!');

if ml_field_set_to_value(opts, 'DecompTol')
    ml_assert_nonnegscalar(opts.DecompTol, 'opts.DecompTol');
else
    opts.DecompTol = log(n) * eps;
end

if ml_field_set_to_value(opts, 'infdecopts')
    assert(isa(opts.infdecopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.infdecopts has to be a struct!');
else
    opts.infdecopts = struct();
end

if ml_field_set_to_value(opts, 'stabdecopts')
    assert(isa(opts.stabdecopts, 'struct'), ...
        'MORLAB:data', ...
        'The parameter opts.stabdecopts has to be a struct!');
else
    opts.stabdecopts = struct();
end

% Additional check of dimensions for avoiding unnecessary calls.
if ml_field_set_to_value(opts.infdecopts, 'Dimension')
    ml_assert_integer(opts.infdecopts.Dimension, ...
        'opts.infdecopts.Dimension');
else
    opts.infdecopts.Dimension = -1;
end

if ml_field_set_to_value(opts.stabdecopts, 'Dimension')
    ml_assert_integer(opts.stabdecopts.Dimension, ...
        'opts.stabdecopts.Dimension');
else
    opts.stabdecopts.Dimension = -1;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION: INFINITE PART.                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation of deflating subspaces using the matrix disk function.
if opts.infdecopts.Dimension == 0
    ninf        = 0;
    infoINFDISK = [];
else
    svdE  = svd(sys.E);
    alpha = svdE(find(svdE > svdE(1) * opts.DecompTol, 1, 'last')) / ...
        (1.01 * norm(sys.A, 'fro'));
    
    if isempty(alpha)
        alpha = 1.0;
    end
    
    [Aspace, Espace, infoINFDISK] = ml_disk(sys.E, alpha * sys.A, ...
        opts.infdecopts);
    [Q, Z, ninf]                  = ml_getqz(sys.A, sys.E, Aspace, ...
        opts.infdecopts);
end

% Block diagonalization of the system matrices.
if (ninf > 0) && (ninf < n)
    opts.infdecopts.Dimension = n - ninf;
    [U, V, nf]                = ml_getqz(sys.A, sys.E, Espace, ...
        opts.infdecopts);
    
    sys.Ainf = U(: , nf+1:end)' * sys.A * Z(: , 1:ninf);
    sys.A    = Q(: , ninf+1:end)' * sys.A * V(: , 1:nf);
    
    sys.Einf = U(: , nf+1:end)' * sys.E * Z(: , 1:ninf);
    sys.E    = Q(: , ninf+1:end)' * sys.E * V(: , 1:nf);
    
    sys.Binf = U(: , nf+1:end)' * sys.B;
    sys.B    = Q(: , ninf+1:end)' * sys.B;
    
    sys.Cinf = sys.C * Z(: , 1:ninf);
    sys.C    = sys.C * V(: , 1:nf);
elseif ninf == n
    sys.Ainf = sys.A;
    sys.Einf = sys.E;
    sys.Binf = sys.B;
    sys.Cinf = sys.C;
    
    [sys.A, sys.B, sys.C, sys.E] = deal([]);
    
    nf = 0;
else
    [sys.Ainf, sys.Binf, sys.Cinf, sys.Einf] = deal([]);
    
    nf = n;
end

n = n - ninf;


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DECOMPOSITION: ANTI-STABLE PART.                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computation of deflating subspaces using the matrix disk function.
if (opts.stabdecopts.Dimension == 0) || (n == 0)
    nu           = 0;
    infoSTABDISK = [];
else
    [Aspace, Espace, infoSTABDISK] = ml_disk(sys.A - sys.E, ...
        sys.A + sys.E, opts.stabdecopts);
    
    [Q, Z, nu] = ml_getqz(sys.A, sys.E, Aspace, opts.stabdecopts);
end

% Block diagonalization of the system matrices.
if (nu > 0) && (nu < n)
    opts.stabdecopts.Dimension = n - nu;
    [U, V, nf]                 = ml_getqz(sys.A, sys.E, Espace, ...
        opts.stabdecopts);
    
    sys.Au = U(: , nf+1:end)' * sys.A * Z(: , 1:nu);
    sys.A  = Q(: , nu+1:end)' * sys.A * V(: , 1:nf);
    
    sys.Eu = U(: , nf+1:end)' * sys.E * Z(: , 1:nu);
    sys.E  = Q(: , nu+1:end)' * sys.E * V(: , 1:nf);
    
    sys.Bu = U(: , nf+1:end)' * sys.B;
    sys.B  = Q(: , nu+1:end)' * sys.B;
    
    sys.Cu = sys.C * Z(: , 1:nu);
    sys.C  = sys.C * V(: , 1:nf);
elseif nu == n
    sys.Au = sys.A;
    sys.Eu = sys.E;
    sys.Bu = sys.B;
    sys.Cu = sys.C;
    
    [sys.A, sys.B, sys.C, sys.E] = deal([]);
else
    [sys.Au, sys.Bu, sys.Cu, sys.Eu] = deal([]);
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ASSIGN INFORMATION.                                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Collection of all information.
info = struct(...
    'infoINFDISK' , infoINFDISK, ...
    'infoSTABDISK', infoSTABDISK, ...
    'Ninf'        , ninf, ...
    'Ns'          , nf, ...
    'Nu'          , nu);
